# #!/data/pmrs/jin/bin/ruby -I /data/pmrs/challis/lib/Atlas-Indel2/trunk/
$:.unshift File.join(File.dirname(__FILE__),'lib')
require 'read.rb'
require 'read_indel.rb'
require 'read_insertion.rb'
require 'read_deletion.rb'
require 'assertion_failure.rb'
# require 'getoptlong'
require 'getoptlong.rb'
require 'ref_seq.rb'
require 'indel.rb'

$DEBUG = false
default_site_level_z_cutoff = -1
default_min_var_reads = 1
default_min_var_ratio = 0.05
default_strand_dir_filter = false
default_min_total_depth=4


class Atlas_indel2
# put read level model coefficients here:
@@read_intercept = -0.72132
@@read_average_nbq = 0.08582
@@read_var_rate = -71.42767
@@site_intercept = -10.40835
@@site_simple_local_entropy =  3.09046
@@site_passed_read_ratio = 3.33967
@@site_frameshift = -2.65220
@@site_alt_depth_ratio = 0.51920


@@read_level_z_cutoff = -1.3
@@e=2.71828182845904523536

	def initialize(infile, ref_filename, outfile, min_var_reads, min_var_ratio, z_cutoff, strand_dir_filter, sample, min_depth_coverage, show_filtered=false)
		if(sample.nil?)
			if(File.basename(infile) =~ /([^\.\/]+)\./)
				@sample = $1
				puts "\"#{@sample}\" will be used as the sample name"
			else 
				STDERR.puts "Failed to find a sample name in the input filename.\nPlease specify a sample name with the \"-s [name]\" argument"
				exit 1
			end
		else
			@sample = sample
		end
		begin
			@ref_file = ref_filename
			@reference = Ref_seq.new(ref_filename)
		rescue
			raise "Error loading the reference sequence from file #{ref_filename}:\n#{$!}"
		end
		begin
			@infile = open_file(infile)
			@outfile = File.open(outfile, 'w')
	#	rescue
	#		raise "There was an error loading the input or output file:\n#{$!}"
		end
		@min_var_reads = min_var_reads
		@min_var_ratio = min_var_ratio
		@z_cutoff = z_cutoff
		@strand_dir_filter = strand_dir_filter
		@show_filtered = show_filtered
		@min_depth_coverage = min_depth_coverage
		if(min_depth_coverage < 4)
			STDERR.puts "The minimum total depth coverage must be at least 4, please specify a higher minimum total depth with the -t argument"
			exit 1
		end
	end


	def self.process_arguments()
		opts = GetoptLong.new(
			["--bam", "-b", GetoptLong::REQUIRED_ARGUMENT],
			["--reference", "-r", GetoptLong::REQUIRED_ARGUMENT ],
			["--outfile","-o", GetoptLong::REQUIRED_ARGUMENT],
			["--sample","-s", GetoptLong::REQUIRED_ARGUMENT],
			["--min-total-depth","-t", GetoptLong::REQUIRED_ARGUMENT],
			["--min-var-reads","-m", GetoptLong::REQUIRED_ARGUMENT],
			["--min-var-ratio","-v", GetoptLong::REQUIRED_ARGUMENT],
			["--z-cutoff","-z", GetoptLong::REQUIRED_ARGUMENT],
			["--strand-dir-filter","-f", GetoptLong::NO_ARGUMENT]
		)

		opt_hash = {}
		begin
			opts.each do |opt, arg|
				opt_hash[opt] = arg
			end
		rescue GetoptLong::InvalidOption
			usage = true
		end

		if(  usage || !(opt_hash.key?("--bam")) || !(opt_hash.key?("--reference")) || !(opt_hash.key?("--outfile")) )
			puts "USAGE:\n	Atlas-Indel2-Illum-Exome.rb -b input_bam -r reference_sequence -o outfile"
			puts "\t\toptional arguments:"
			puts "\t\t-s --sample [by default taken from BAM file name] (the name of the sample)"
			puts "\t\t-z --z-cutoff [-1] (the z-score cutoff for the regression model)"
			puts "\t\t-t --min-total-depth [4] (may not be lower than 4)"
			puts "\t\t-m --min-var-reads [1]"
			puts "\t\t-v --min-var-ratio [0.05]"
			puts "\t\t-f --strand-dir-filter (requires at least one read in each direction, extremely limits sensitivity)"
			exit 0
		end

		return opt_hash
	end




	def process_data
		puts "Progress:"
		print "loading...\r"
		skipped = 0
		skipped_flagged = 0
		processed = 0
		processed_chars = 0
		progress = 0
		sec = 0
		buffered_coor = 0 #how far has the current chromosome been buffered?
		print_header
		chr = nil
		indel_buffer = Hash.new
		depth = Hash.new
		ref_depth = Hash.new
		line = get_line()
		@counter=0
		while(line != :eof)
			begin
				
				read = Read.new(line)
				chr = read.ref if chr == nil
				buffered_coor = read.pos - 1 # one less to allow depth coverage to work correctly
				if(indel_buffer.length > 0 && chr == read.ref) # indels are in the buffer, print if possible
					print_buffer(indel_buffer, buffered_coor, chr, depth, ref_depth)
				elsif(chr != read.ref) #reached end of chromosome, print buffer
					print_buffer(indel_buffer, 1.0/0.0, chr, depth, ref_depth) #pass infinity as the buffered_coor to empty the buffer
					assert(indel_buffer.length == 0, "ERROR: Indel buffer is not empty after printing end of chromosome #{chr}. Buffer length: #{indel_buffer.length}\n#{indel_buffer.to_s}")
					indel_buffer = Hash.new
					chr = read.ref
					depth = Hash.new
					ref_depth = Hash.new
				end
				indels = read.get_indels_with_depth(depth, ref_depth) # 
				# update_depth(read, chr, depth, ref_depth)
				if(indels.length > 0)
					var_rate = read.var_rate
				end
				indels.each do |read_indel|
					process_indel(read_indel, indel_buffer,  var_rate)
				end
				processed += 1
			rescue FormatError
				STDERR.puts $!.to_s + "\n\n" + line + "\n\n"
				skipped += 1
			rescue FlaggedError
				skipped_flagged += 1
				skipped += 1
			# rescue 
			#	puts "There was an unknown error with a line, it will be skipped.  If you see this message many times you should contact the developer.\n\n#{line}\n\n#{$!}" # skipped += 1
			ensure
				if( Time.now.sec != sec && Time.now.sec % 5 == 0)
					indicate_progress(chr, read.pos)
					sec = Time.now.sec
				end
				line = get_line()
			end
		end
		@infile.close
		print_buffer(indel_buffer, 1.0/0.0, chr, depth, ref_depth)  #print whatever is left in the buffer
		assert(indel_buffer.length == 0, "ERROR: Indel buffer is not empty at the end of the last chromosome. Buffer length: #{indel_buffer.length}\n#{indel_buffer.to_s}")
		@outfile.close
		puts "DONE                                      "
		puts "#{skipped} lines skipped "
		puts "#{skipped_flagged} flagged lines skipped"
		puts "#{processed} lines processed"
	end



	private
	

	def update_depth(read, chr, depth, ref_depth)
		(read.pos..read.end_pos).each do |coor|
			if(depth[coor] == nil)
				depth[coor] = 1
				ref_depth[coor] = 0
			else
				depth[coor]+= 1
			end
			ref_depth[coor] += 1 if read.seq[coor-read.pos] == @reference.get_base(chr.to_sym, coor)
		end
	end


	def open_file(infile)
		#begin
			return IO.popen("samtools view -F 1796 #{infile}")
		#rescue
			raise "Error: failed to extract SAM lines from BAM file"
		#end
		
	end


	def print_header
		@outfile.puts '##fileformat=VCFv4.0'
		@outfile.puts "##fileDate=#{Time.now.strftime("%Y%m%d")}"
		@outfile.puts "##source=Atlas-Indel2-Illum-Exome"
		@outfile.puts "##reference=#{@ref_file}"
		@outfile.puts "##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">"
		@outfile.puts "##FORMAT=<ID=GQ,Number=1,Type=Integer,Description=\"Genotype Quality\">"
		@outfile.puts "##FORMAT=<ID=DP,Number=1,Type=Integer,Description=\"Read Depth\">"
		@outfile.puts "##FORMAT=<ID=RR,Number=1,Type=Integer,Description=\"Reference Read Depth\">"
		@outfile.puts "##FORMAT=<ID=VR,Number=1,Type=Integer,Description=\"Major Variant Read Depth\">"
		@outfile.puts "#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT	#{@sample}"
	end


	def indicate_progress(chr, coor)
		print "Chromosome:#{chr}  Coor:#{coor}                           \r"
		STDOUT.flush
		STDERR.flush
		@outfile.flush
	end


	def process_indel(read_indel, indel_buffer, var_rate)
		z=nil
		z = read_level_model(read_indel.avnbq, var_rate)
		indel_key = read_indel.to_s
		if( indel_buffer[indel_key] == nil)
			indel_buffer[indel_key] = Indel.new(read_indel, z, @@read_level_z_cutoff)
		else
			indel_buffer[indel_key].add_read_indel(z, read_indel)
		end
	end



	def print_buffer(buffer, buffer_coor, chr, depth, ref_depth)
		# print "writing output from buffer..."
		# STDOUT.flush
		i=buffer.length
		prev_indel = nil
		keys = buffer.keys
		keys.sort! if i > 1
		keys.each do |key|
			indel = buffer[key]
			if(!prev_indel.nil?)
			  	if( indel.read_indel.start_pos == prev_indel.read_indel.start_pos) # check for multiple indels at the same site (we only support one per site)
					if(prev_indel.read_count < indel.read_count)
						prev_indel = indel
					end
					buffer.delete(key)
					next
				else
					print_indel(prev_indel, chr,  depth, ref_depth)
					prev_indel = nil
				end
			end
			if(indel.read_indel.start_pos < buffer_coor)
				buffer.delete(key) # clean out indel buffer
				prev_indel = indel
			end
		end
		print_indel(prev_indel, chr, depth, ref_depth) unless prev_indel.nil?
		depth.delete_if { |coor, depth| coor < buffer_coor } #clean out the depth buffer
		ref_depth.delete_if { |coor, depth| coor < buffer_coor } #clean out the ref_depth buffer
		# print "#{i-buffer.length} lines printed\n"
		# STDOUT.flush
	end


	def print_indel(indel, chr, depth, ref_depth)
		raise "ERROR: missing total depth information for indel #{indel.read_indel.to_s} -- #{indel.read_indel.start_pos()}" if depth[indel.read_indel.start_pos()] == nil
		#assert( chr == key.split(":")[0].to_sym, "ERROR: tried to print #{indel} out on chromosome #{chr}")
		#assert( key.split(/[:DI]/)[1].to_i == indel.read_indel.start_pos, "ERROR: tried to print #{indel} out at wrong coordinate")
		if(@show_filtered)
			filter, qual = filter(indel, depth)
			if(filter == :PASS)
				@outfile.puts "#{chr}\t#{indel.read_indel.start_pos}\t.\t#{indel.read_indel.ref_seq}\t#{indel.read_indel.alt_seq}\t#{qual}\tPASS\t.\tGT:DP:VR:RR:GQ\t1/.:#{depth[indel.read_indel.start_pos()]}:#{indel.read_count}:#{ref_depth[indel.read_indel.start_pos()]}:."

			else
				@outfile.puts "#{chr}\t#{indel.read_indel.start_pos}\t.\t#{indel.read_indel.ref_seq}\t#{indel.read_indel.alt_seq}\t#{qual}\t#{filter.join(";")}\t.\tGT:DP:VR:RR:GQ\t1/.:#{depth[indel.read_indel.start_pos()]}:#{indel.read_count}:#{ref_depth[indel.read_indel.start_pos()]}:."
			end
		else
			qual = passes_filters(indel, depth)
			if(qual)
				@outfile.puts "#{chr}\t#{indel.read_indel.start_pos}\t.\t#{indel.read_indel.ref_seq}\t#{indel.read_indel.alt_seq}\t#{qual}\tPASS\t.\tGT:DP:VR:RR:GQ\t1/.:#{depth[indel.read_indel.start_pos()]}:#{indel.read_count}:#{ref_depth[indel.read_indel.start_pos()]}:."
			end
		end
	end




	def read_level_model(average_nbq, var_rate)
		return @@read_intercept + @@read_average_nbq*average_nbq + @@read_var_rate*var_rate
	end


	def get_line
		begin
			line = @infile.readline
			line.chomp!
		rescue EOFError => err
			return :eof
		end
		return line
	end

	def passes_filters(indel, depth)
		return false if indel.read_count < @min_var_reads
		total_depth = depth[indel.read_indel.start_pos()]
		return false if total_depth < @min_depth_coverage
		var_ratio = indel.read_count.to_f / total_depth.to_f
		return false if var_ratio < @min_var_ratio
		return false if @strand_dir_filter && indel.fails_strand_dir_filt
		z = run_site_model(indel, total_depth)
		p=1/(1+@@e**-z)
		return false if z < @z_cutoff
		return 100 if 1-p < 0.00000000001
		return -10* Math.log10(1-p).round
	end

	def filter(indel, depth)
		filters = Array.new
		filters.push "var#{@min_var_rads}" if indel.read_count < @min_var_reads
		filters.push 'sDir' if @strand_dir_filter && indel.fails_strand_dir_filt
		total_depth = depth[indel.read_indel.start_pos()]
		filters.push "dp#{@min_depth_coverage}" if total_depth < @min_depth_coverage
		var_ratio = indel.read_count.to_f / total_depth.to_f
		filters.push "vr#{@min_var_ratio}"  if var_ratio < @min_var_ratio
		z = run_site_model(indel, total_depth)
		p=1/(1+@@e**-z)
		filters.push "atlas#{@z_cutoff}" if p < @z_cutoff
		return :PASS if filters.length == 0
		return filters, -10* Math.log10(1-p).round
	end



	def run_site_model(indel, total_depth)
		#return 1/(1+@@e**-(
		passed_read_ratio = indel.num_passed_reads.to_f/indel.read_count.to_f
		return @@site_intercept + @@site_simple_local_entropy*indel.simple_local_entropy +
				  @@site_passed_read_ratio*passed_read_ratio + @@site_frameshift*indel.frameshift +
				  @@site_alt_depth_ratio*((indel.read_count**2)/total_depth)
	end

end

# MAIN ##########################
if __FILE__ == $0
	begin
		opt_hash = Atlas_indel2.process_arguments()
		sample = opt_hash["--sample"]
		if(opt_hash["--min-var-reads"].nil?)
			min_var_reads = default_min_var_reads
		else
			min_var_reads = opt_hash["--min-var-reads"].to_i
			puts "using minimum variant read requirement of #{min_var_reads}"
		end
		if( opt_hash["--z-cutoff"].nil? )
			z_cutoff = default_site_level_z_cutoff
		else
			z_cutoff = opt_hash["--z-cutoff"].to_f
			puts "using z-score cutoff of #{z_cutoff}"
		end
		if( opt_hash["--min-var-ratio"].nil? )
			min_var_ratio = default_min_var_ratio 
		else
			min_var_ratio = opt_hash["--min-var-ratio"].to_f
			puts "using minimum variant ratio requirement of #{min_var_ratio}"
		end
		if( opt_hash["--min-total-depth"].nil? )
			min_total_depth = default_min_total_depth
		else
			min_total_depth = opt_hash["--min-total-depth"].to_i
			puts "using minimum total depth requirement of #{min_total_depth}"
		end
		if( opt_hash["--strand-dir-filter"].nil? )
			strand_dir_filter = default_strand_dir_filter 
		else
			strand_dir_filter = opt_hash["strand-dir-filter"]
			puts "using strand direction filter"
		end
		atlas = Atlas_indel2.new(opt_hash["--bam"], opt_hash["--reference"], opt_hash["--outfile"], min_var_reads, min_var_ratio, z_cutoff, strand_dir_filter, sample, min_total_depth, false)
		atlas.process_data
	
		print "\nAtlas-INDEL2 Finished\n"
	rescue EORefError
		puts "FATAL ERROR: Either your reference sequence is not sorted the same as your bam file, or it is missing a chromosome listed in your BAM file."
		exit 1
	rescue
		puts "UNKOWN FATAL ERROR: Please contact the developers so they can address this issue (challis@bcm.edu, fyu@bcm.edu).\nPlease include the following details:\n$!"
		exit 1
	end

end
